/*************************************************************************\
* Copyright (c) 2008 - 2015 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * Code the for Master Timing IOC that generates tick-reset broadcast
 * packets and synchronizes the APS Event timing system to NTP.
 * It also provides a service used by 3.13 soft-timing slave IOCs to
 * obtain the current time on demand.
 */

#include <string.h>
#include <stdlib.h>
#include <sntpcLib.h>

#include <epicsStdioRedirect.h>
#include <epicsGeneralTime.h>
#include <generalTimeSup.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <osiSock.h>
#include <envDefs.h>

#include "drvTS.h"
#include "devApsEr.h"


extern char* sysBootLine;

/* Symbols from erTimeProvider.h */

extern int ErInhibitBroadcastProvider;
extern int ErEventTickRate(void);
extern int ErTimeProviderPriority(void);
extern int ErPutEventRaw(epicsTimeStamp *pDest, int eventNumber);
extern int ErGetEventRaw(epicsTimeStamp *pDest, int eventNumber);

/* Module macros */

#define TS_SYNC_RATE_SEC 10
#define ER_EVENT_RESET_TICK		0x7d	/* Reset the tick counter */
#define NTP_TIMEOUT (4 * sysClkRateGet())

/* Global variables */

int erTimeMasterDebug = 0;
int tmSyncReportThreshold = 0;
int tmNTPReportThreshold = 0;

/* Module variables */

static epicsTimeStamp lastNTPsync;

/* Forward definitions */

int tmSyncToNTP(void);


/*
 * tmStampServer - A service on the master timing IOC that listens for
 * current time requests and sync requests from slaves. Only IOCs that
 * are 3.13.10 drvTS soft timing slaves normally use this protocol, so
 * this server could be removed once all 3.13.10 IOCs have been upgraded.
 */

static void tmStampServer(void *dummy)
{
    osiSockAddr sin;
    int sock = epicsSocketCreate(AF_INET, SOCK_DGRAM, 0);

    if (sock < 0) {
        fprintf(stderr, "tmStampServer: Socket creation failed\n");
        return;
    }

    memset(&sin, 0, sizeof(sin));
    sin.ia.sin_family = AF_INET;
    sin.ia.sin_addr.s_addr = htonl(INADDR_ANY);
    sin.ia.sin_port = htons(TS_MASTER_PORT);

    if (bind(sock, &sin.sa, sizeof(sin.ia)) < 0) {
        perror("tmStampServer: socket bind failed");
        return;
    }

    for(;;) {
        TSstampTrans stran;
        struct sockaddr fs;
        epicsTimeStamp ts;
        int fl = sizeof(fs);
        int mlen = recvfrom(sock, (char *) &stran, sizeof(stran), 0, &fs, &fl);
        int type;
        int status;

        if (mlen < 0) {
            perror("tmStampServer: recvfrom() failed");
            continue;
        }

        if ((mlen < 2 * sizeof(epicsUInt32)) ||
            (mlen > sizeof(TSstampTrans))) {
            fprintf(stderr, "tmStampServer: Unexpected %d byte packet seen\n",
                mlen);
            continue;
        }

        if (ntohl(stran.magic) != TS_MAGIC) {
            fprintf(stderr, "tmStampServer: Bad packet received\n");
            continue;
        }

        type = ntohl(stran.type);
        switch (type) {
        case TS_time_request:
            if (erTimeMasterDebug >= 5) {
                char buf[30];

                sockAddrToDottedIP(&fs, buf, sizeof buf);
                printf("tmStampServer: Slave at %s requested time\n", buf);
            }

            if (mlen < 40)
                mlen = 40;

            status = epicsTimeGetEvent(&ts, ER_EVENT_RESET_TICK);
            if (status) {
                fprintf(stderr, "tmStampServer: epicsTimeGetEvent() failed\n");
                break;
            }
            stran.master_time.tv_sec = htonl(ts.secPastEpoch);
            stran.master_time.tv_nsec = htonl(ts.nsec);
            if (erTimeMasterDebug >= 8) {
                char buf[35];

                epicsTimeToStrftime(buf, sizeof buf,
                    "%Y-%m-%d %H:%M:%S.%09f", &ts);
                printf("\tmaster_time = %s\n", buf);
            }

            status = epicsTimeGetCurrent(&ts);
            if (status) {
                fprintf(stderr, "tmStampServer: epicsTimeGetCurrent() failed\n");
                break;
            }
            stran.current_time.tv_sec = htonl(ts.secPastEpoch);
            stran.current_time.tv_nsec = htonl(ts.nsec);
            if (erTimeMasterDebug >= 8) {
                char buf[35];

                epicsTimeToStrftime(buf, sizeof buf,
                    "%Y-%m-%d %H:%M:%S.%09f", &ts);
                printf("\tcurrent_time = %s\n", buf);
            }

            stran.sync_rate = htonl(TS_SYNC_RATE_SEC);
            stran.clock_hz = htonl(ErEventTickRate());

            status = sendto(sock, (char *) &stran, mlen, 0, &fs, fl);
            if (status < 0) {
                perror("tmStampServer: sendto() failed");
            }
            break;

        case TS_sync_request:
            if (erTimeMasterDebug >= 2) {
                char buf[30];

                sockAddrToDottedIP(&fs, buf, sizeof buf);
                printf("tmStampServer: Slave at %s requested NTP sync\n", buf);
            }
            tmSyncToNTP();
            break;

        default: {
                char buf[30];

                sockAddrToDottedIP(&fs, buf, sizeof buf);
                fprintf(stderr, "tmStampServer: "
                    "Unknown request type %d from slave at %s\n",
                    type, buf);
            }
        }
    }
    close(sock);
}

int tmStartStampServer(void)
{
    epicsThreadId id;

    id = epicsThreadCreate("tm_stamp", epicsThreadPriorityHigh,
        epicsThreadGetStackSize(epicsThreadStackSmall), tmStampServer, 0);
    return id ? 0 : -1;
}


/*
 *  Network Socket Support Routines
 */

/*
 * tmGetBroadcastAddr - use libCom's osiSock.h API to get a a broadcast
 * address for this IOC. Assumes there is only one network interface, or
 * that the first interface found is the right one for broadcasting to.
 */

static int tmGetBroadcastAddr(int sock, osiSockAddr* sin)
{
    ELLLIST tmpList;
    osiSockAddrNode *pNode;

    ellInit(&tmpList);
    osiSockDiscoverBroadcastAddresses(&tmpList, sock, sin);

    /* Use first-found */
    pNode = (osiSockAddrNode *) ellFirst(&tmpList);
    if (!pNode) {
        fprintf(stderr, "tmGetBroadcastAddr: No broadcast address!\n");
        return -1;
    }

    memcpy(sin, &pNode->addr, sizeof(struct sockaddr));

    /* Free address list items */
    ellFree(&tmpList);
    return 0;
}


/*
 * tmCreateBroadcastSocket - creates a socket for sending broadcasts, sets
 * sockaddr to the broadcast address.
 */

static int tmCreateBroadcastSocket(osiSockAddr* sin)
{
    int on = 1;
    int sock = epicsSocketCreate(AF_INET, SOCK_DGRAM, 0);

    if (sock < 0) {
        perror("tmCreateBroadcastSocket: Socket creation failed");
        return -1;
    }

    memset(&sin->ia, 0, sizeof(sin->ia));
    sin->ia.sin_port = htons(0);
    sin->ia.sin_family = AF_INET;
    sin->ia.sin_addr.s_addr = htonl(INADDR_ANY);

    setsockopt(sock, SOL_SOCKET, SO_BROADCAST, (char *) &on, sizeof(on));

    if (bind(sock, &sin->sa, sizeof(sin->ia)) < 0) {
        perror("tmCreateBroadcastSocket: Socket bind failed");
        close(sock);
        return -1;
    }

    if (tmGetBroadcastAddr(sock, sin) < 0)
        return -1;

    return sock;
}


/*
 * tmCreateLocalSocket - creates a socket for sending datagrams, sets
 * sockaddr to the local address.
 */

static int tmCreateLocalSocket(osiSockAddr* sin)
{
    int sock = epicsSocketCreate(AF_INET, SOCK_DGRAM, 0);

    if (sock < 0) {
        perror("tmCreateLocalSocket: Socket creation failed");
        return -1;
    }

    *sin = osiLocalAddr(sock);

    return sock;
}


/*
 *  Tick Counter Reset Subroutines
 */

static TSstampTrans resetTrans;
static int broadcastSock, localSock;
static osiSockAddr broadcastAddr, localAddr;
static const char *ntpServer = NULL;
static int tmValid;

void tmTickResetInit(void)
{
    if (broadcastSock > 0)  /* Already initialized */
        return;

    broadcastSock = tmCreateBroadcastSocket(&broadcastAddr);
    localSock = tmCreateLocalSocket(&localAddr);
    if (broadcastSock < 0 || localSock < 0) {
        return;
    }

    broadcastAddr.ia.sin_port = htons(TS_SLAVE_PORT);
    localAddr.ia.sin_port = htons(TS_SLAVE_PORT);

    resetTrans.magic     = htonl(TS_MAGIC);
    resetTrans.type      = htonl(TS_sync_msg);
    resetTrans.sync_rate = htonl(TS_SYNC_RATE_SEC);
    resetTrans.clock_hz  = htonl(ErEventTickRate());

    ntpServer = envGetConfigParamPtr(&EPICS_TS_NTP_INET);
    if (!ntpServer) { /* use the boot host */
        BOOT_PARAMS bootParms;
        static char host_addr[BOOT_ADDR_LEN];

        bootStringToStruct(sysBootLine, &bootParms);
        /* bootParms.had = host IP address */
        strncpy(host_addr, bootParms.had, BOOT_ADDR_LEN);
        ntpServer = host_addr;
    }

    /* Just in case user didn't run ErTimeMasterConfigure() */
    ErInhibitBroadcastProvider = 1;
}

int tmTickReset(void)
{
    epicsTimeStamp tRESET, tNOW;
    epicsUInt32 mask = 0;

    if (broadcastSock <= 0)
        return -1;

    if (ErGetEventRaw(&tRESET, ER_EVENT_RESET_TICK)) {
        printf("tmTickReset: ErGetEventRaw(ER_EVENT_RESET_TICK) failed\n");
        return -1;
    }

    if (!tmValid) {
        /* Initial or resync reset tick processing */
        struct timespec tsNTP;
        epicsUInt32 t0, t1;
        double latency;

        /* The period between the reset event and now includes the following:
         *    1. Interrupt latency, ErIrqHandler()
         *    2. Processing of any previous events in the EVR100 FIFO
         *    3. Task switch from scanIoRequest() to cbHigh task
         *    4. Processing of any previous records in the cbHigh queue
         *       and their after-effects, output & forward links etc.
         *    5. Record processing of Mt:EV:ep_event_125
         *    6. Record processing of Mt:TickResetSubX and subTickReset()
         *
         * The time read from NTP gets corrected for that delay and for
         * the network latencies involved in contacting the NTP server.
         */

        ErGetTicks(0, &t0);
        if (sntpcTimeGet((char *) ntpServer, NTP_TIMEOUT, &tsNTP)) {
            perror("tmTickReset: sntpcTimeGet() failed");
            return -1;
        }
        ErGetTicks(0, &t1);

        /* How long did it take to fetch the time over the network? */
        latency = (t1 - t0) / ErEventTickRate();

        if (tRESET.secPastEpoch < 600) {
            /* First time through we set tRESET to tsNTP - latency/2 */
            epicsTimeFromTimespec(&tRESET, &tsNTP);
            epicsTimeAddSeconds(&tRESET, latency / -2.0);
            if (erTimeMasterDebug >= 4) {
                char buf[35];

                epicsTimeToStrftime(buf, sizeof buf,
                    "%Y-%m-%d %H:%M:%S.%09f", &tRESET);
                printf("tmTickReset: Initialized ErEvent clock to %s\n",
                    buf);
            }
        }
        else {
            epicsTimeStamp tNTP;
            double correction;

            if (epicsTimeGetCurrent(&tNOW)) {
                fprintf(stderr, "tmTickReset: epicsTimeGetCurrent() failed\n");
                return -1;
            }
            epicsTimeFromTimespec(&tNTP, &tsNTP);

            /* Add half the latency, since tNOW was measured *after* tNTP */
            epicsTimeAddSeconds(&tNTP, latency / 2);

            /* Adjust tRESET by how much tNOW is offset from tNTP */
            correction = epicsTimeDiffInSeconds(&tNTP, &tNOW);
            epicsTimeAddSeconds(&tRESET, correction);
            if (erTimeMasterDebug >= 4) {
                char buf[35];
                double sinceNtp = epicsTimeDiffInSeconds(&tNTP, &lastNTPsync);

                printf("tmTickReset: Adding NTP correction %.9f seconds\n",
                    correction);
                epicsTimeToStrftime(buf, sizeof buf,
                    "%Y-%m-%d %H:%M:%S.%09f", &tRESET);
                printf("tmTickReset: Reset tick is now %s\n",
                    buf);

                printf("tmTickReset: %.6f seconds since last NTP correction\n"
                    "    Correction was %g seconds/second\n",
                    sinceNtp, correction / sinceNtp);
            }
            lastNTPsync = tNTP;
        }

        ErPutEventRaw(&tRESET, 0);
        tmValid = 1;
        mask |= TS_STAMP_FROM_UNIX;
    }
    else if (epicsTimeGetCurrent(&tNOW)) {
        fprintf(stderr, "tmTickReset: epicsTimeGetCurrent() failed\n");
        tNOW.secPastEpoch = tNOW.nsec = 0;
    }

    resetTrans.master_time.tv_sec = htonl(tRESET.secPastEpoch);
    resetTrans.master_time.tv_nsec = htonl(tRESET.nsec);
    resetTrans.current_time.tv_sec = htonl(tNOW.secPastEpoch);
    resetTrans.current_time.tv_nsec = htonl(tNOW.nsec);
    resetTrans.mask = htonl(mask);
    resetTrans.syncReportThreshold = htonl(tmSyncReportThreshold);
    resetTrans.syncUnixReportThreshold = htonl(tmNTPReportThreshold);

    if (sendto(broadcastSock, (char *) &resetTrans, sizeof(resetTrans), 0,
            &broadcastAddr.sa, sizeof(broadcastAddr.ia)) < 0) {
        perror("tmTickReset: sendto(broadcast) failed");
        return -1;
    }

    if (sendto(localSock, (char *) &resetTrans, sizeof(resetTrans), 0,
            &localAddr.sa, sizeof(localAddr.ia)) < 0) {
        perror("tmTickReset: sendto(me) failed");
        return -1;
    }

    return 0;
}


/*
 *  NTP Synchronization Subroutines
 */

void tmSyncInit(void)
{
    /* Nothing? */
}

int tmSyncToNTP(void)
{
    tmValid = 0;    /* Next tmTickReset() will re-synchronize */

    return 0;
}


/*
 *  Timing Master Configuration command
 */

void ErTimeMasterConfigure(int eventRate, int syncReport, int ntpReport)
{
    ErInhibitBroadcastProvider = 1;
    ErTimeProviderConfigure(0, eventRate);

    tmSyncReportThreshold = syncReport;
    tmNTPReportThreshold  = ntpReport;

    if (tmStartStampServer() < 0)
        perror("tmTickResetInit: tmStartStampServer failed");
}


