/*************************************************************************\
* Copyright (c) 2008 - 2015 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* EPICS BASE is distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * APS Event Receiver Time Provider
 * APS Event Broadcast Time Provdier
 */
#include <epicsStdioRedirect.h>
#include <epicsGeneralTime.h>
#include <epicsVersion.h>
#include <generalTimeSup.h>
#include <epicsStdlib.h>
#include <cantProceed.h>
#include <string.h>
#include <math.h>
#include <iocsh.h>

/* This is needed for vxWorks 6.8 to prevent an obnoxious compiler warning */
#define _VSB_CONFIG_FILE <../lib/h/config/vsbConfig.h>

#include <time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>

#include <epicsInterrupt.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <epicsExport.h>
#include <errlog.h>
#include <osiSock.h>
#include "devApsEr.h"
#include "drvTS.h"

/*
 * Assume that broadcaster is dead if nothing received in this time
 */
#define BROADCAST_TIMEOUT_SECONDS   120

/*
 * Assume that event sender is dead if this many broadcasts arrive without events
 */
#define MISSED_RESET_TICKS_EVENT_LIMIT    2

/*
 * Event system parameters
 */
#define ER_EVENT_TICK_RATE 1000
#define ER_1MHZ_TICK_RATE 1000000
#define ER_DEFAULT_TICK_RATE ER_EVENT_TICK_RATE
#define ER_RESET_TICKS_EVENT 0x7d

/*
 * Global variables
 */
int ErInhibitBroadcastProvider = 0;
int erTimeProviderDebug = 0;

/*
 * Interrupt-level communication
 */
static volatile int eventsGood;
static volatile int consecutiveEventReceiverErrorCount;
static volatile int missedResetTicksEvent;
static volatile unsigned int eventCounter;
static volatile int lastEvent;
static volatile unsigned int errorCounter;
static volatile int lastError;

/*
 * Internal variables
 */
static int             broadcastsGood;
static unsigned int    broadcastCounter;
static int             eventTimePriority;
static int             eventTickRate;
static int             nsecPerEventTick;
static epicsTimeStamp *eventTable;
static int             eventTableCapacity;
static epicsTimeStamp  timeAtLastBroadcast;
static int             systemTicksAtLastBroadcast;
static int             systemTickRate;
static int             nsecPerSystemTick;
static TSeventFunc     otherEventHandler;
static TSerrorFunc     otherErrorHandler;

/*
 * Register event receiver methods
 */
static struct drvTSops tsOps, *devTSops;
void drvTsRegisterDeviceSupport(struct drvTSops *ops)
{
    if (devTSops)
        printf("APS event receiver time provider -- duplicate event providers!\n");
    devTSops = ops;
}


/*
 * Return the current time using event-receiver provider.
 * Must be callable from interrupt context!
 */
static int
erGetTime(epicsTimeStamp *pDest)
{
    epicsTimeStamp ts;
    epicsUInt32 ticks;
    int key;

    key = epicsInterruptLock();
    if (!broadcastsGood
     || !eventsGood
     || (missedResetTicksEvent >= MISSED_RESET_TICKS_EVENT_LIMIT)) {
        epicsInterruptUnlock(key);
        return epicsTimeERROR;
    }
    ErGetTicks(0, &ticks);
    *pDest = eventTable[ER_RESET_TICKS_EVENT];
    epicsInterruptUnlock(key);

    if (pDest->secPastEpoch < 60*60*24*365)
        return epicsTimeERROR;

    ts.secPastEpoch = ticks / eventTickRate;
    ts.nsec = (ticks - (ts.secPastEpoch * eventTickRate)) * nsecPerEventTick;
    pDest->secPastEpoch += ts.secPastEpoch;
    pDest->nsec += ts.nsec;
    if (pDest->nsec >= 1000000000) {
        pDest->secPastEpoch++;
        pDest->nsec -= 1000000000;
    }
    return epicsTimeOK;
}

/*
 * Return the time when an event occurred.
 */
static int
erGetEventTime(epicsTimeStamp *pDest, int eventNumber)
{
    int key;

    if ((eventNumber <= 0)
     || (eventNumber >= eventTableCapacity))
        return epicsTimeERROR;
    key = epicsInterruptLock();
    if (!broadcastsGood || !eventsGood) {
        epicsInterruptUnlock(key);
        return epicsTimeERROR;
    }
    *pDest = eventTable[eventNumber];
    epicsInterruptUnlock(key);
    return epicsTimeOK;
}

/*
 * Return the current time using event-broadcast provider.
 */
static int
ebGetTime(epicsTimeStamp *pDest)
{
    epicsTimeStamp base, interval;
    epicsUInt32 lastTicks, currentTicks, intervalTicks;
    int key;

    if (ErInhibitBroadcastProvider)
        return epicsTimeERROR;

    key = epicsInterruptLock();
    if (!broadcastsGood) {
        epicsInterruptUnlock(key);
        return epicsTimeERROR;
    }
    base = timeAtLastBroadcast;
    currentTicks = osdTickGet();
    lastTicks = systemTicksAtLastBroadcast;
    epicsInterruptUnlock(key);
    /*
     * Note that the following calculation works
     * even when currentTicks has wrapped around
     */
    intervalTicks = currentTicks - lastTicks;
    interval.secPastEpoch = intervalTicks / systemTickRate;
    interval.nsec = (intervalTicks - interval.secPastEpoch * systemTickRate) * nsecPerSystemTick;
    if (interval.secPastEpoch > BROADCAST_TIMEOUT_SECONDS) {
        broadcastsGood = 0;
        return epicsTimeERROR;
    }
    pDest->secPastEpoch = base.secPastEpoch + interval.secPastEpoch;
    pDest->nsec = base.nsec + interval.nsec;
    if (pDest->nsec >= 1000000000) {
        pDest->secPastEpoch++;
        pDest->nsec -= 1000000000;
    }
    return epicsTimeOK;
}

/*
 * Set/get the time for a event. These routines used by erTimeMaster.c
 */

int ErPutEventRaw(epicsTimeStamp *pts, int eventNumber)
{
    int key;

    if ((eventNumber <= 0)
     || (eventNumber >= eventTableCapacity))
        return epicsTimeERROR;
    key = epicsInterruptLock();
    eventTable[eventNumber] = *pts;
    epicsInterruptUnlock(key);
    return epicsTimeOK;
}

int ErGetEventRaw(epicsTimeStamp *pDest, int eventNumber)
{
    int key;

    if ((eventNumber <= 0)
     || (eventNumber >= eventTableCapacity))
        return epicsTimeERROR;
    key = epicsInterruptLock();
    *pDest = eventTable[eventNumber];
    epicsInterruptUnlock(key);
    return epicsTimeOK;
}


int ErEventShow(int all)
{
    int i;

    for (i=1; i<eventTableCapacity; i++) {
        epicsTimeStamp ts;
        int key = epicsInterruptLock();

        ts = eventTable[i];
        epicsInterruptUnlock(key);
        if (all || ts.secPastEpoch || ts.nsec) {
            char cbuf[35];

            epicsTimeToStrftime(cbuf, sizeof cbuf, "%Y-%m-%d %H:%M:%S.%09f", &ts);
            printf("  %3d: %s\n", i, cbuf);
        }
    }
    return 0;
}


/*
 * Receive events from event system -- update the event table 
 * Note: called at interrupt level!
 */
static void
erEventHandler(const int Card, const int EventNum, const epicsUInt32 ticks)
{
    epicsTimeStamp ts, *st;

    if (eventsGood && (EventNum == ER_RESET_TICKS_EVENT)) {
        missedResetTicksEvent = 0;
        if (broadcastsGood && (--broadcastsGood == 0))
            epicsInterruptContextMessage(
               "***apsEvent system: No packets, or no time change, from master\n");
        else
            if (erTimeProviderDebug >= 8)
                epicsInterruptContextMessage(
                    "erEventHandler: tick reset event\n");
    }
    eventCounter++;
    lastEvent = EventNum;
    eventsGood = 1;
    consecutiveEventReceiverErrorCount = 0;

    /* calculate a time stamp from the event tick count */
    ts.secPastEpoch = ticks / eventTickRate;
    ts.nsec = (ticks - (ts.secPastEpoch * eventTickRate)) * nsecPerEventTick;
    /* obtain a copy of the last sync time */
    st = &eventTable[ER_RESET_TICKS_EVENT];
    /* update the event table */
    eventTable[EventNum].secPastEpoch = st->secPastEpoch + ts.secPastEpoch;
    eventTable[EventNum].nsec = st->nsec + ts.nsec;
    /* adjust seconds if needed */
    if (eventTable[EventNum].nsec >= 1000000000) {
        eventTable[EventNum].secPastEpoch++;
        eventTable[EventNum].nsec -= 1000000000;
    }
    if (otherEventHandler)
        otherEventHandler(Card, EventNum, ticks);
}

/*
 * Receive error reports from event system
 * Note: called at interrupt level!
 */
static void
errorHandler(const int Card, const int ErrorNum)
{
    eventsGood = 0;
    errorCounter++;
    lastError = ErrorNum;
    if (consecutiveEventReceiverErrorCount < 4) {
        consecutiveEventReceiverErrorCount++;
        switch(ErrorNum)
        {
        case 1:
            epicsInterruptContextMessage(
               "***apsEvent system: TAXI violation\n\n");
            break;
        case 2:
            epicsInterruptContextMessage(
               "***apsEvent system: lost heartbeat\n\n");
            break;
        case 3:
            epicsInterruptContextMessage(
                "***apsEvent system: lost events\n\n");
            break;
        default:
            epicsInterruptContextMessage(
                "***apsEvent system: unknown error from event system\n\n");
        }
    }
    if (otherErrorHandler)
        otherErrorHandler(Card, ErrorNum);
}

/*
 * Listen for time stamps from master.
 */
static void
timeBroadcastMonitor(void *arg)
{
    epicsTimeStamp oldMasterTime = {0, 0};
    struct sockaddr_in sin;
    SOCKET sock;

    memset(&sin, 0, sizeof(sin));
    sin.sin_port = htons(TS_SLAVE_PORT);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);

    sock = epicsSocketCreate(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        errlogPrintf("erTimeProvider terminating, can't create socket: %s\n",
            strerror(SOCKERRNO));
        return;
    }

    if (bind(sock, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
        errlogPrintf("erTimeProvider terminating, can't bind socket: %s\n",
            strerror(SOCKERRNO));
        close(sock);
        return;
    }

    for (;;) {
        epicsTimeStamp masterTime, broadcastTime, myTime;
        struct sockaddr fs;
        TSstampTrans stran;
        osiSocklen_t fl = sizeof(fs);
        int i = recvfrom(sock, (char *)&stran, sizeof(stran), 0, &fs, &fl);
        int key, newServer;

        if (i < 0) {
            errlogPrintf("erTimeProvider: Recv failed: %s\n",
                strerror(SOCKERRNO));
            continue;
        }

        if (i == (sizeof(stran) - 3 * sizeof(epicsUInt32))) {
            /* R3.13 master doesn't send entire packet */
            stran.mask = htonl(0);
            stran.syncReportThreshold = htonl(0);
            stran.syncUnixReportThreshold = htonl(0);
            newServer = 0;
        }
        else if (i != sizeof(stran)) {
            errlogPrintf("erTimeProvider: received %d, expected %u.\n", i, (unsigned int)sizeof(stran));
            continue;
        }
        else
            newServer = 1;

        if (ntohl(stran.magic) != TS_MAGIC) {
            errlogPrintf("erTimeProvider: invalid packet received.\n");
            continue;
        }

        if (erTimeProviderDebug >= 1000) {
            eventsGood = 0;   /* Simulate faulty event system */
            if (erTimeProviderDebug >= 2000)
                continue;     /* Simulate faulty master */
        }

        masterTime.secPastEpoch = ntohl(stran.master_time.tv_sec);
        masterTime.nsec         = ntohl(stran.master_time.tv_nsec);

        if (newServer && stran.current_time.tv_sec) {
            broadcastTime.secPastEpoch = ntohl(stran.current_time.tv_sec);
            broadcastTime.nsec         = ntohl(stran.current_time.tv_nsec);
        }
        else {
            broadcastTime = masterTime;
        }

        if (erTimeProviderDebug >= 2) {
            char cbuf[35];

            epicsTimeToStrftime(cbuf, sizeof cbuf, "%Y-%m-%d %H:%M:%S.%09f",
                &masterTime);
            printf("erTimeProvider: Sync received, master_time = %s\n", cbuf);
        }

        if (epicsTimeEqual(&masterTime, &oldMasterTime))
            continue;       /* Master seems to have locked up */

        oldMasterTime = masterTime;
        broadcastsGood = 2;
        broadcastCounter++;

        key = epicsInterruptLock();
        systemTicksAtLastBroadcast = osdTickGet();
        timeAtLastBroadcast = broadcastTime;
        myTime = eventTable[ER_RESET_TICKS_EVENT];
        missedResetTicksEvent++;
        epicsInterruptUnlock(key);

        if (!eventsGood)
            continue;

        if (missedResetTicksEvent == MISSED_RESET_TICKS_EVENT_LIMIT)
            errlogPrintf("erTimeProvider: RESET_TICK events missing\n");

        if (missedResetTicksEvent >= MISSED_RESET_TICKS_EVENT_LIMIT)
            continue;

        if (epicsTimeNotEqual(&myTime, &masterTime)) {
            epicsUInt32 mask;
            epicsUInt32 syncReportThreshold;
            epicsUInt32 syncUnixReportThreshold;
            double reportThreshold;
            double unixReportThreshold;
            double timeDiff;
            static int firstUpdate = 1;

            timeDiff = epicsTimeDiffInSeconds(&masterTime, &myTime);

            mask = ntohl(stran.mask);
            syncReportThreshold = ntohl(stran.syncReportThreshold);
            syncUnixReportThreshold = ntohl(stran.syncUnixReportThreshold);
            reportThreshold = syncReportThreshold/1000.0;
            unixReportThreshold = syncUnixReportThreshold/1000.0;

            key = epicsInterruptLock();
            eventTable[ER_RESET_TICKS_EVENT] = masterTime;
            epicsInterruptUnlock(key);

            if (firstUpdate) {
                firstUpdate = 0;
            }
            else if ((mask & TS_STAMP_FROM_UNIX) &&
                (fabs(timeDiff) >= unixReportThreshold)) {
                char cbuf[35];

                epicsTimeToStrftime(cbuf, sizeof cbuf,
                    "%Y-%m-%d %H:%M:%S.%09f", &masterTime);
                errlogPrintf("\nerTimeProvider: Event Time Sync from NTP\n"
                    "    Adjusted by %-.9f seconds to %s\n",
                    timeDiff, cbuf);
            }
            else if (!(mask & TS_STAMP_FROM_UNIX) &&
                (fabs(timeDiff) >= reportThreshold)) {
                char cbuf[35];

                epicsTimeToStrftime(cbuf, sizeof cbuf,
                    "%Y-%m-%d %H:%M:%S.%09f", &masterTime);
                errlogPrintf("\nerTimeProvider: Event Time Synchronized\n"
                    "    Adjusted by %-.9f seconds to %s\n",
                    timeDiff, cbuf);
            }
        }
        else if (erTimeProviderDebug >= 5) {
            printf("erTimeProvider: Perfect time sync with master\n");
        }
    }
}

int ErEventTickRate(void)
{
    return eventTickRate;
}

int ErTimeProviderPriority(void)
{
    return eventTimePriority;
}

int ErTimeReport(int level)
{
    char buf[35];

    if (!systemTickRate) {
        printf("erTimeProvider not configured\n");
        return 0;
    }

    printf("erTimeProvider was configured as follows:\n");
    printf("  Time provider priority = %d\n", eventTimePriority);
    printf("  Tick counter clock rate = %d (J4)\n", eventTickRate);

    printf("\nAPS Time Broadcast provider is %s\n",
        ErInhibitBroadcastProvider ? "Inhibited" : "Enabled");

    if (level == 0) {
        printf("\nAPS Event time provider status is %s\n",
            eventsGood ? "Good" : "Bad");
        if (!ErInhibitBroadcastProvider)
            printf("\nAPS Time Broadcast provider status is %s\n",
                broadcastsGood ? "Good" : "Bad");
    }
    else {
        printf("\nEvent receiver status:\n");
        printf("  eventsGood = %d (%s)\n", eventsGood,
            eventsGood ? "Good" : "Bad");
        printf("  eventCounter = %u\n", eventCounter);
        printf("  lastEvent = %d\n", lastEvent);
        printf("  errorCounter = %u\n", errorCounter);
        printf("  lastError = %d\n", lastError);

        if (level > 1) {
            epicsUInt32 ticks;

            ErGetTicks(0, &ticks);
            printf("  tick count = %u\n", ticks);
        }

        printf("\nBroadcast monitor status:\n");
        printf("  broadcastsGood = %d (%s)\n", broadcastsGood,
            broadcastsGood ? "Good" : "Bad");
        printf("  broadcastCounter = %u\n", broadcastCounter);
        printf("  missedResetTicksEvent = %d\n", missedResetTicksEvent);
        epicsTimeToStrftime(buf, sizeof buf,
            "%Y-%m-%d %H:%M:%S.%09f", &timeAtLastBroadcast);
        printf("  timeAtLastBroadcast = %s\n", buf);
    }

    return 0;
}

void ErTimeProviderConfigure(int priority, int eventRate)
{
    static int firstTime = 1;

    if (!firstTime)
        return;
    firstTime = 0;
    systemTickRate = osdTickRateGet();
    nsecPerSystemTick = 1000000000 / systemTickRate;

    if (priority == 0)
        priority = 70;
    eventTimePriority = priority;

    switch (eventRate) {
    case 0:
        eventRate = ER_DEFAULT_TICK_RATE;
    case ER_EVENT_TICK_RATE:
    case ER_1MHZ_TICK_RATE:
        break;

    default:
        printf("ErTimeProviderConfigure: eventRate may be 0, %d or %d only\n",
            ER_EVENT_TICK_RATE, ER_1MHZ_TICK_RATE);
        return;
    }
    eventTickRate = eventRate;
    nsecPerEventTick = 1000000000/eventRate;

    if (devTSops != NULL) {
        tsOps = *devTSops;
        if ((tsOps.TShaveReceiver != NULL)
         && ((eventTableCapacity = tsOps.TShaveReceiver(0)) > 0)) {
            /* allocate the event table */
            eventTable = callocMustSucceed(eventTableCapacity, sizeof(*eventTable), "ErTimeProviderConfigure");

            otherEventHandler = tsOps.TSregisterEventHandler(0, erEventHandler);
            otherErrorHandler = tsOps.TSregisterErrorHandler(0, errorHandler);

            generalTimeRegisterCurrentProvider("APS Event Receiver", priority, erGetTime);
            generalTimeAddIntCurrentProvider("APS Event Receiver", priority, erGetTime);

            generalTimeRegisterEventProvider("APS Event Receiver", priority, erGetEventTime);
            generalTimeAddIntEventProvider("APS Event Receiver", priority, erGetEventTime);
        }
    }

    if (!ErInhibitBroadcastProvider)
        generalTimeCurrentTpRegister("APS Time Broadcast", priority+5, ebGetTime);

    if (epicsThreadCreate("timeBroadcastMonitor", epicsThreadPriorityHigh,
            epicsThreadGetStackSize(epicsThreadStackSmall),
            timeBroadcastMonitor, NULL) == 0) {
        errlogPrintf("Failed to start time-broadcast monitor client\n");
    }
}

static const iocshArg ErTimeProviderConfigureArg0 = {"priority", iocshArgInt};
static const iocshArg ErTimeProviderConfigureArg1 = {"event_rate", iocshArgInt};
static const iocshArg *const ErTimeProviderConfigureArgs[] = { &ErTimeProviderConfigureArg0 };
static const iocshFuncDef ErTimeProviderConfigureDef = {"ErTimeProviderConfigure",1,ErTimeProviderConfigureArgs};
static void ErTimeProviderConfigureCall(const iocshArgBuf * args)
{
    ErTimeProviderConfigure(args[0].ival, args[1].ival);
}

static void
erTimeProviderRegister(void)
{
    static int firstTime = 1;
    if (firstTime) {
        firstTime = 0;
        iocshRegister(&ErTimeProviderConfigureDef,ErTimeProviderConfigureCall);
    }
}
epicsExportAddress(int,erTimeProviderDebug);
epicsExportRegistrar(erTimeProviderRegister);
